@@ -239,12 +239,13 @@ static duk_ret_t wrapped_compile_execute(duk_context *ctx, void *udata) {
 
 	if (src_data != NULL && src_len >= 2 && src_data[0] == (char) 0xff) {
 		/* Bytecode. */
-		duk_push_lstring(ctx, src_data, src_len);
-		duk_to_buffer(ctx, -1, NULL);
+		void *buf;
+		buf = duk_push_fixed_buffer(ctx, src_len);
+		memcpy(buf, (const void *) src_data, src_len);
 		duk_load_function(ctx);
 	} else {
 		/* Source code. */
-		comp_flags = 0;
+		comp_flags = DUK_COMPILE_SHEBANG;
 		duk_compile_lstring_filename(ctx, comp_flags, src_data, src_len);
 	}
 
@@ -289,7 +290,7 @@ static duk_ret_t wrapped_compile_execute(duk_context *ctx, void *udata) {
 
 #if 0
 	/* Manual test for bytecode dump/load cycle: dump and load before
-	 * execution.  Enable manually, then run "make qecmatest" for a
+	 * execution.  Enable manually, then run "make ecmatest" for a
 	 * reasonably good coverage of different functions and programs.
 	 */
 	duk_dump_function(ctx);
@@ -1122,7 +1123,7 @@ static duk_context *create_duktape_heap(int alloc_provider, int debugger, int aj
 	if (!ctx) {
 		fprintf(stderr, "Failed to create Duktape heap\n");
 		fflush(stderr);
-		exit(-1);
+		exit(1);
 	}
 
 #if defined(DUK_CMDLINE_AJSHEAP)
