@@ -130,7 +130,7 @@ def create_targz(dstfile, filelist):
 def cstring(x):
     return '"' + x + '"'  # good enough for now
 
-# DUK_VERSION is grepped from duk_api_public.h.in: it is needed for the
+# DUK_VERSION is grepped from duktape.h.in: it is needed for the
 # public API and we want to avoid defining it in two places.
 def get_duk_version(apiheader_filename):
     r = re.compile(r'^#define\s+DUK_VERSION\s+(.*?)L?\s*$')
@@ -255,6 +255,8 @@ def main():
     parser.add_option('--verbose', dest='verbose', action='store_true', default=False, help='Show verbose debug messages')
 
     (opts, args) = parser.parse_args()
+    if len(args) > 0:
+        raise Exception('unexpected arguments: %r' % args)
 
     if opts.obsolete_builtin_metadata is not None:
         raise Exception('--user-builtin-metadata has been removed, use --builtin-file instead')
@@ -310,7 +312,7 @@ def main():
             duk_dist_meta = json.loads(f.read())
 
     duk_version, duk_major, duk_minor, duk_patch, duk_version_formatted = \
-        get_duk_version(os.path.join(srcdir, 'duk_api_public.h.in'))
+        get_duk_version(os.path.join(srcdir, 'duktape.h.in'))
 
     git_commit = None
     git_branch = None
@@ -430,10 +432,13 @@ def main():
         'duk_bi_symbol.c',
         'duk_bi_thread.c',
         'duk_bi_thrower.c',
+        'duk_dblunion.h',
         'duk_debug_fixedbuffer.c',
         'duk_debug.h',
         'duk_debug_macros.c',
         'duk_debug_vsnprintf.c',
+        'duk_debugger.c',
+        'duk_debugger.h',
         'duk_error_augment.c',
         'duk_error.h',
         'duk_error_longjmp.c',
@@ -445,11 +450,14 @@ def main():
         'duk_hbuffer_alloc.c',
         'duk_hbuffer.h',
         'duk_hbuffer_ops.c',
+        'duk_hbufobj.h',
+        'duk_hbufobj_misc.c',
         'duk_hcompfunc.h',
         'duk_heap_alloc.c',
         'duk_heap.h',
         'duk_heap_hashstring.c',
         'duk_heaphdr.h',
+        'duk_heap_finalize.c',
         'duk_heap_markandsweep.c',
         'duk_heap_memory.c',
         'duk_heap_misc.c',
@@ -460,7 +468,6 @@ def main():
         'duk_hobject_alloc.c',
         'duk_hobject_class.c',
         'duk_hobject_enum.c',
-        'duk_hobject_finalizer.c',
         'duk_hobject.h',
         'duk_hobject_misc.c',
         'duk_hobject_pc2line.c',
@@ -472,10 +479,7 @@ def main():
         'duk_hthread.h',
         'duk_hthread_misc.c',
         'duk_hthread_stacks.c',
-        'duk_hbufobj.h',
-        'duk_hbufobj_misc.c',
-        'duk_debugger.c',
-        'duk_debugger.h',
+        'duk_henv.h',
         'duk_internal.h',
         'duk_jmpbuf.h',
         'duk_exception.h',
@@ -493,6 +497,7 @@ def main():
         'duk_lexer.h',
         'duk_numconv.c',
         'duk_numconv.h',
+        'duk_refcount.h',
         'duk_regexp_compiler.c',
         'duk_regexp_executor.c',
         'duk_regexp.h',
@@ -505,7 +510,6 @@ def main():
         'duk_util_bitencoder.c',
         'duk_util.h',
         'duk_util_hashbytes.c',
-        'duk_util_hashprime.c',
         'duk_util_misc.c',
         'duk_util_tinyrandom.c',
         'duk_util_bufwriter.c',
@@ -608,8 +612,6 @@ def main():
         '@DUK_SINGLE_FILE@': '#define DUK_SINGLE_FILE',
         '@LICENSE_TXT@': read_file(os.path.join(tempdir, 'LICENSE.txt.tmp'), strip_last_nl=True),
         '@AUTHORS_RST@': read_file(os.path.join(tempdir, 'AUTHORS.rst.tmp'), strip_last_nl=True),
-        '@DUK_API_PUBLIC_H@': read_file(os.path.join(srcdir, 'duk_api_public.h.in'), strip_last_nl=True),
-        '@DUK_DBLUNION_H@': read_file(os.path.join(srcdir, 'duk_dblunion.h.in'), strip_last_nl=True),
         '@DUK_VERSION_FORMATTED@': duk_version_formatted,
         '@GIT_COMMIT@': git_commit,
         '@GIT_COMMIT_CSTRING@': git_commit_cstring,
@@ -896,7 +898,6 @@ def main():
             'duk_error_macros.c',
             'duk_unicode_support.c',
             'duk_util_misc.c',
-            'duk_util_hashprime.c',
             'duk_hobject_class.c'
         ]
 
